import jax
import numpy as np
import sys
import time

from pyRDDLGym.Planner import JaxConfigManager
from pyRDDLGym.Core.Compiler.RDDLDecompiler import RDDLDecompiler
# from pyRDDLGym.Core.Jax.JaxRDDLModelError import JaxRDDLModelError


def print_parameterized_exprs(planner):
    model_params = planner.compiled.model_params
    print(f'model_params = {model_params}')
    ids = planner.compiled.get_ids_of_parameterized_expressions()
    for _id in ids:
        expr = planner.compiled.traced.lookup(_id)
        print(f'\nid = {_id}:\n' + RDDLDecompiler().decompile_expr(expr))
    
    
def slp_train(planner, budget, **train_args):
    step = train_args['step']
    del train_args['step']
    print('\n' + 'training plan:')
    for i, callback in enumerate(planner.optimize(**train_args, step=1)):
        if i == 0:
            elapsed = 0
            starttime = time.time()
        else:
            elapsed = time.time() - starttime
        
        if i % step == 0:
            print('[{:.4f} s] step={} train_return={:.6f} test_return={:.6f} best_return={:.6f}'.format(
                elapsed,
                str(callback['iteration']).rjust(4),
                callback['train_return'],
                callback['test_return'],
                callback['best_return']))
        if elapsed >= budget:
            print('ran out of time!')
            break
    params = callback['best_params']
    
    # key = jax.random.PRNGKey(42)
    # error = JaxRDDLModelError(planner.rddl, planner.test_policy, 
    #                           batch_size=64, logic=planner.logic)
    # error.summarize(key, params)
    # error.sensitivity(key, params)
    return params


def slp_no_replan(env, trials, timeout, timeout_ps, save):
    myEnv, planner, _, _, train_args, (dom, inst) = JaxConfigManager.get(f'{env}.cfg')
    key = train_args['key']    
    
    rewards = np.zeros((myEnv.horizon, trials))
    for trial in range(trials):
        print('\n' + '*' * 30 + '\n' + f'starting trial {trial + 1}\n' + '*' * 30)
        train_args['key'] = key
        params = slp_train(planner, timeout, **train_args)
        
        total_reward = 0
        state = myEnv.reset()
        for step in range(myEnv.horizon):
            subs = myEnv.sampler.subs
            key, subkey = jax.random.split(key)
            action = planner.get_action(subkey, params, step, subs)
            next_state, reward, done, _ = myEnv.step(action)
            total_reward += reward 
            rewards[step, trial] = reward
            
            print()
            print('step       = {}'.format(step))
            print('state      = {}'.format(state))
            print('action     = {}'.format(action))
            print('next state = {}'.format(next_state))
            print('reward     = {}'.format(reward))
            state = next_state
            if done:
                break
        print(f'episode ended with reward {total_reward}')
        
    myEnv.close()
    if save:
        np.savetxt(f'{dom}_{inst}_slp.csv', rewards, delimiter=',')

    
def slp_replan(env, trials, timeout, timeout_ps, save):
    myEnv, planner, _, _, train_args, (dom, inst) = JaxConfigManager.get(f'{env}.cfg')
    key = train_args['key']
    
    rewards = np.zeros((myEnv.horizon, trials))
    for trial in range(trials):
        print('\n' + '*' * 30 + '\n' + f'starting trial {trial + 1}\n' + '*' * 30)
        total_reward = 0
        state = myEnv.reset() 
        starttime = time.time()
        train_args['guess'] = None
        for step in range(myEnv.horizon):
            currtime = time.time()
            elapsed = currtime - starttime
            
            if elapsed < timeout:
                subs = myEnv.sampler.subs
                params = slp_train(planner,
                                   budget=min(timeout - elapsed, timeout_ps),
                                   subs=subs,
                                   **train_args)
                key, subkey = jax.random.split(key)
                action = planner.get_action(subkey, params, 0, subs)
                train_args['guess'] = planner.plan.guess_next_epoch(params)
            else:
                print('ran out of time!')
                action = {}
            
            next_state, reward, done, _ = myEnv.step(action)
            total_reward += reward 
            rewards[step, trial] = reward
            
            print()
            print(f'elapsed    = {elapsed} s')
            print(f'step       = {step}')
            print(f'state      = {state}')
            print(f'action     = {action}')
            print(f'next state = {next_state}')
            print(f'reward     = {reward}')
            state = next_state
            if done: 
                break
        print(f'episode ended with reward {total_reward}')
        
    myEnv.close()
    if save:
        np.savetxt(f'{dom}_{inst}_mpc.csv', rewards, delimiter=',')

    
def main(env, replan, trials, timeout, timeout_ps, save):
    if replan:
        slp_replan(env, trials, timeout, timeout_ps, save)
    else: 
        slp_no_replan(env, trials, timeout, timeout_ps, save)
    
        
if __name__ == "__main__":
    if len(sys.argv) < 6:
        env, trials, timeout, timeout_ps, save = 'Wildfire', 1, 60 * 2, 1, False
    else:
        env, trials, timeout, timeout_ps, save = sys.argv[1:6]
        trials = int(trials)
        timeout = int(timeout)
        timeout_ps = int(timeout_ps)
        save = save == 'True' or save == True
    replan = env.endswith('replan')
    main(env, replan, trials, timeout, timeout_ps, save) 
    



from ast import literal_eval
import configparser
import jax
import jax.nn.initializers as initializers
import optax
import os
from typing import Dict
import warnings

from pyRDDLGym.Core.Env.RDDLEnv import RDDLEnv
from pyRDDLGym.Core.Jax import JaxRDDLBackpropPlanner
from pyRDDLGym.Core.Jax import JaxRDDLLogic
from pyRDDLGym.Examples.ExampleManager import ExampleManager


def load_config_file(path: str):
    path = os.path.join(os.path.dirname(os.path.abspath(__file__)), path)
    config = configparser.RawConfigParser()
    config.optionxform = str 
    config.read(path)
    args = {k: literal_eval(v) 
            for section in config.sections()
            for (k, v) in config.items(section)}
    return config, args


def read_config_sections(config, args):
    env_args = {k: args[k] for (k, _) in config.items('Environment')} 
    model_args = {k: args[k] for (k, _) in config.items('Model')}
    planner_args = {k: args[k] for (k, _) in config.items('Optimizer')}
    train_args = {k: args[k] for (k, _) in config.items('Training')}
    return env_args, model_args, planner_args, train_args


def load_rddl_files(check_external, domain_name, inst_name):
    try: 
        # try to read from external rddlrepository  
        if not check_external:
            raise Exception
        warnings.warn(f'reading {domain_name} from rddlrepository...', stacklevel=2)
        from rddlrepository.Manager.RDDLRepoManager import RDDLRepoManager
        manager = RDDLRepoManager()
        EnvInfo = manager.get_problem(domain_name)
    except: 
        # default to embedded RDDL manager
        warnings.warn(f'failed to read from rddlrepository, '
                      f'reading {domain_name} from Examples...', stacklevel=2)
        EnvInfo = ExampleManager.GetEnvInfo(domain_name)        
    domain = EnvInfo.get_domain()
    instance = EnvInfo.get_instance(inst_name)
    return domain, instance


def get(path: str, 
        new_env_kwargs={}, new_logic_kwargs={}, 
        new_plan_kwargs={}, new_planner_kwargs={}) -> Dict[str, object]:
    
    # load the config file
    config, args = load_config_file(path)
    env_args, model_args, planner_args, train_args = read_config_sections(config, args)
    
    # read the environment settings
    check_external = env_args.pop('check_rddlrepository', False)
    domain_name = env_args['domain']
    inst_name = env_args['instance']
    env_args['domain'], env_args['instance'] = load_rddl_files(
        check_external, domain_name, inst_name)
    env_args.update(new_env_kwargs)
    myEnv = RDDLEnv(**env_args)
    planner_args['rddl'] = myEnv.model
    
    # read the model settings
    tnorm_name = model_args['tnorm']
    tnorm_kwargs = model_args['tnorm_kwargs']
    logic_name = model_args['logic']
    logic_kwargs = model_args['logic_kwargs']
    logic_kwargs['tnorm'] = getattr(JaxRDDLLogic, tnorm_name)(**tnorm_kwargs)
    logic_kwargs.update(new_logic_kwargs)
    planner_args['logic'] = getattr(JaxRDDLLogic, logic_name)(**logic_kwargs)
    
    # read the optimizer settings
    plan_method = planner_args.pop('method')
    plan_kwargs = planner_args.pop('method_kwargs', {})  
    
    if 'initializer' in plan_kwargs:  # weight initialization
        init_name = plan_kwargs['initializer']
        init_class = getattr(initializers, init_name)
        init_kwargs = plan_kwargs.pop('initializer_kwargs', {})
        try: 
            plan_kwargs['initializer'] = init_class(**init_kwargs)
        except:
            warnings.warn(f'ignoring arguments for initializer <{init_name}>',
                          stacklevel=2)
            plan_kwargs['initializer'] = init_class
               
    if 'activation' in plan_kwargs:  # activation function
        plan_kwargs['activation'] = getattr(jax.nn, plan_kwargs['activation'])
    
    plan_kwargs.update(new_plan_kwargs)
    planner_args['plan'] = getattr(JaxRDDLBackpropPlanner, plan_method)(**plan_kwargs)
    planner_args['optimizer'] = getattr(optax, planner_args['optimizer'])
    planner_args.update(new_planner_kwargs)
    planner = JaxRDDLBackpropPlanner.JaxRDDLBackpropPlanner(**planner_args)
    
    # read the training settings
    train_args['key'] = jax.random.PRNGKey(train_args['key'])
    
    return myEnv, planner, planner_args, plan_kwargs, train_args, \
        (domain_name, inst_name)
    



import abc

from itertools import cycle


class RDDLEnvSeeder(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def Next(self) -> float:
        pass

class RDDLEnvSeederFibonacci(RDDLEnvSeeder):
    def __init__(self, a1=None):
        self.an = 0
        if a1 is None:
            a1 = 1
        self.anext = a1

    def Next(self) -> float:
        next = self.anext
        self.anext = self.an + next
        self.an = next
        return next


class RDDLEnvSeederCyclic(RDDLEnvSeeder):
    def __init__(self, seed_list: list=None):
        self.list = seed_list
        self.iterator = None
        if self.list is not None:
            if len(self.list) > 0:
                self.iterator = cycle(self.list)

    def Next(self):
        if self.iterator is not None:
            return next(self.iterator)
        else:
            return None




from pyRDDLGym.Visualizer.visualize_dbn import RDDL2Graph


def test_dbn_visualization():
    domains = [
        'cartpole discrete', 'cartpole continuous', 'marsrover', 
        # 'mountaincar',    # Takes too long to create DBN graph for this
        'powergen', 'racecar', 'uavcontinuous', 
        'uavdiscrete', 'uavmixed', 'wildfire', 'supplychain', 
        'propdbn'
    ]
    
    for domain in domains:
        r2g = RDDL2Graph(
            domain=domain,
            instance=0,
            directed=True,
            strict_grouping=True
        )
        r2g.save_dbn(file_name=domain)


if __name__ == "__main__":
    test_dbn_visualization()



import sys
import os

from pyRDDLGym import RDDLEnv
from pyRDDLGym import ExampleManager
from pyRDDLGym.Policies.Agents import RandomAgent
from pyRDDLGym.Visualizer.MovieGenerator import MovieGenerator
from pyRDDLGym.Visualizer.TextViz import TextVisualizer

def main(env, inst, method_name=None, episodes=1):
    print(f'preparing to launch instance {inst} of domain {env}...')
    
    # get the environment info
    # ExampleManager.RebuildExamples()
    EnvInfo = ExampleManager.GetEnvInfo(env)
    
    # set up the environment class, choose instance 0 because every example has at least one example instance
    log = False if method_name is None else True
    myEnv = RDDLEnv.RDDLEnv(domain=EnvInfo.get_domain(), 
                            instance=EnvInfo.get_instance(inst),
                            enforce_action_constraints=False,
                            debug=False,
                            log=log,
                            simlogname=method_name)
    myEnv.seed(42)
    
    # set up the environment visualizer
    # frames_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Visualizer', 'Frames')
    myEnv.set_visualizer(EnvInfo.get_visualizer())#, movie_gen=MovieGenerator(frames_path, env, 200), movie_per_episode=True)
    # myEnv.set_visualizer(None, movie_gen=MovieGenerator(frames_path, env, 200),
    #                      movie_per_episode=True)
    
    # set up an example aget
    agent = RandomAgent(action_space=myEnv.action_space, 
                        num_actions=myEnv.numConcurrentActions,
                        seed=42)

    for episode in range(episodes):
        total_reward = 0
        state = myEnv.reset()
        for step in range(myEnv.horizon):
            myEnv.render()
            action = agent.sample_action()
            next_state, reward, done, info = myEnv.step(action)
            total_reward += reward
            print()
            print(f'step       = {step}')
            print(f'state      = {state}')
            print(f'action     = {action}')
            print(f'next state = {next_state}')
            print(f'reward     = {reward}')
            state = next_state
            if done:
                break
        print(f'episode {episode} ended with reward {total_reward}')
    
    myEnv.close()


if __name__ == "__main__":
    args = sys.argv
    method_name = None
    episodes = 1
    if len(args) < 3:
        env, inst = 'HVAC', '0'
    elif len(args) < 4:
        env, inst = args[1:3]
    elif len(args) < 5:
        env, inst, method_name = args[1:4]
    else:
        env, inst, method_name, episodes = args[1:5]
        try:
            episodes = int(episodes)
        except:
            raise ValueError('episodes argument must be an integer, received:' + episodes)
    main(env, inst, method_name, episodes)



import numpy as np
import warnings

from pyRDDLGym.Core.ErrorHandling.RDDLException import print_stack_trace

from pyRDDLGym.Core.Simulator.RDDLSimulator import RDDLSimulator


class RDDLConstraints:

    def __init__(self, simulator: RDDLSimulator, max_bound: float=np.inf) -> None:
        self.sim = simulator
        self.rddl = simulator.rddl
        self.BigM = max_bound
        self.epsilon = 0.001
        
        self._bounds = {}
        for (var, vtype) in self.rddl.variable_types.items():
            if vtype in {'state-fluent', 'observ-fluent', 'action-fluent'}:
                ptypes = self.rddl.param_types[var]
                for gname in self.rddl.ground_names(var, ptypes):
                    self._bounds[gname] = [-self.BigM, +self.BigM]

        # actions and states bounds extraction for gym's action and state spaces
        # currently supports only linear inequality constraints
        for precond in self.rddl.preconditions:
            self._parse_bounds(precond, [], self.rddl.actions)
            
        for invariant in self.rddl.invariants:
            self._parse_bounds(invariant, [], self.rddl.states)

        for (name, bounds) in self._bounds.items():
            RDDLSimulator._check_bounds(*bounds, f'Variable <{name}>', bounds)
            
        # log bounds to file
        if simulator.logger is not None:
            bounds_info = '\n\t'.join(
                f'{k}: {v}' for (k, v) in self._bounds.items())
            message = (f'computed simulation bounds:\n' 
                       f'\t{bounds_info}\n')
            simulator.logger.log(message)
        
    def _parse_bounds(self, expr, objects, search_vars):
        etype, op = expr.etype
        
        if etype == 'aggregation' and op == 'forall':
            * pvars, arg = expr.args
            new_objects = objects + [pvar for (_, pvar) in pvars]
            self._parse_bounds(arg, new_objects, search_vars)
            
        elif etype == 'boolean' and op == '^':
            for arg in expr.args:
                self._parse_bounds(arg, objects, search_vars)
                
        elif etype == 'relational':
            var, lim, loc, active = self._parse_bounds_relational(
                expr, objects, search_vars)
            if var is not None and loc is not None: 
                if objects:
                    ptypes = [ptype for (_, ptype) in objects]
                    variations = self.rddl.variations(ptypes)
                    lims = np.ravel(lim, order='C')
                    for (args, lim) in zip(variations, lims):
                        active_args = [args[i] for i in active]
                        key = self.rddl.ground_name(var, active_args)
                        self._update_bound(key, loc, lim)
                else:
                    self._update_bound(var, loc, lim)
    
    def _update_bound(self, key, loc, lim):
        if loc == 1:
            if self._bounds[key][loc] > lim:
                self._bounds[key][loc] = lim
        else:
            if self._bounds[key][loc] < lim:
                self._bounds[key][loc] = lim
        
    def _parse_bounds_relational(self, expr, objects, search_vars):
        left, right = expr.args    
        _, op = expr.etype
        is_left_pvar = left.is_pvariable_expression() and left.args[0] in search_vars
        is_right_pvar = right.is_pvariable_expression() and right.args[0] in search_vars
        
        if (is_left_pvar and is_right_pvar) or op not in ['<=', '<', '>=', '>']:
            warnings.warn(
                f'Constraint does not have a structure of '
                f'<action or state fluent> <op> <rhs>, where:' 
                    f'\n<op> is one of {{<=, <, >=, >}}'
                    f'\n<rhs> is a deterministic function of '
                    f'non-fluents or constants only.\n' + 
                    print_stack_trace(expr))
            return None, 0.0, None, []
            
        elif not is_left_pvar and not is_right_pvar:
            return None, 0.0, None, []
        
        else:
            if is_left_pvar:
                var, args = left.args
                const_expr = right
            else:
                var, args = right.args
                const_expr = left
            if args is None:
                args = []
                
            if not self.rddl.is_non_fluent_expression(const_expr):
                warnings.warn(
                    f'Bound must be a deterministic function of '
                    f'non-fluents or constants only.\n' + 
                    print_stack_trace(const_expr))
                return None, 0.0, None, []
            
            const = self.sim._sample(const_expr, self.sim.subs)
            eps, loc = self._get_op_code(op, is_left_pvar)
            lim = const + eps
            
            arg_to_index = {obj[0]: i for (i, obj) in enumerate(objects)}
            active = [arg_to_index[arg] for arg in args if arg in arg_to_index]

            return var, lim, loc, active
            
    def _get_op_code(self, op, is_right):
        eps = 0.0
        if is_right:
            if op in ['<=', '<']:
                loc = 1
                if op == '<':
                    eps = -self.epsilon
            elif op in ['>=', '>']:
                loc = 0
                if op == '>':
                    eps = self.epsilon
        else:
            if op in ['<=', '<']:
                loc = 0
                if op == '<':
                    eps = self.epsilon
            elif op in ['>=', '>']:
                loc = 1
                if op == '>':
                    eps = -self.epsilon
        return eps, loc

    @property
    def bounds(self):
        return self._bounds

    @bounds.setter
    def bounds(self, value):
        self._bounds = value


import sys

from pyRDDLGym.Core.Jax.JaxParameterTuning import JaxParameterTuningSLP
from pyRDDLGym.Core.Jax.JaxParameterTuning import JaxParameterTuningSLPReplan
from pyRDDLGym.Core.Jax.JaxParameterTuning import JaxParameterTuningDRP
from pyRDDLGym.Planner import JaxConfigManager


def tune(env, method, trials, timeout, timeout_ps, iters, workers):
    myEnv, _, planner_args, plan_args, train_args, _ = JaxConfigManager.get(f'{env}.cfg')
    key = train_args['key']    
    
    if method == 'slp':
        tuning = JaxParameterTuningSLP(
            env=myEnv,
            max_train_epochs=train_args['epochs'],
            timeout_episode=timeout,
            verbose=True,
            print_step=None, #train_args['step'],
            planner_kwargs=planner_args,
            plan_kwargs=plan_args,
            num_workers=workers,
            gp_iters=iters)
        best = tuning.tune(key, 'gp_slp')
        
    elif method == 'replan':
        tuning = JaxParameterTuningSLPReplan(
            env=myEnv,
            max_train_epochs=train_args['epochs'],
            timeout_episode=timeout,
            timeout_epoch=timeout_ps,
            eval_trials=trials,
            verbose=True,
            print_step=None, #train_args['step'],
            planner_kwargs=planner_args,
            plan_kwargs=plan_args,
            num_workers=workers,
            gp_iters=iters)
        best = tuning.tune(key, 'gp_replan')
        
    elif method == 'drp':
        tuning = JaxParameterTuningDRP(
            env=myEnv,
            max_train_epochs=train_args['epochs'],
            timeout_episode=timeout,
            verbose=True,
            print_step=None, #train_args['step'],
            planner_kwargs=planner_args,
            plan_kwargs=plan_args,
            num_workers=workers,
            gp_iters=iters)
        best = tuning.tune(key, 'gp_drp')
    
    print(f'best parameters found = {best}')


if __name__ == "__main__":
    if len(sys.argv) < 7:
        env, trials, timeout, timeout_ps, iters, workers = 'CartPole_continuous_drp', 1, 10, 1, 20, 4
    else:
        env, trials, timeout, timeout_ps, iters, workers = sys.argv[1:7]
        trials = int(trials)
        timeout = int(timeout)
        timeout_ps = int(timeout_ps)
        iters = int(iters)
        workers = int(workers)
    if env.endswith('replan'):
        method = 'replan'
    elif env.endswith('drp'):
        method = 'drp'
    else:
        method = 'slp'
    tune(env, method, trials, timeout, timeout_ps, iters, workers) 
    



import copy
import gym
from gym.spaces import Discrete, Dict, Box
import numpy as np
import pygame
import os

from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLInvalidNumberOfArgumentsError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLTypeError, RDDLLogFolderError

from pyRDDLGym.Core.Compiler.RDDLLiftedModel import RDDLLiftedModel
from pyRDDLGym.Core.Debug.Logger import Logger, SimLogger
from pyRDDLGym.Core.Env.RDDLConstraints import RDDLConstraints
from pyRDDLGym.Core.Env.RDDLEnvSeeder import RDDLEnvSeederFibonacci as RDDLSeeder
from pyRDDLGym.Core.Parser.parser import RDDLParser
from pyRDDLGym.Core.Parser.RDDLReader import RDDLReader
from pyRDDLGym.Core.Simulator.RDDLSimulator import RDDLSimulator
from pyRDDLGym.Visualizer.TextViz import TextVisualizer


class RDDLEnv(gym.Env):
    
    def __init__(self, domain: str,
                 instance: str=None,
                 enforce_action_constraints: bool=False,
                 enforce_action_count_non_bool: bool=True,
                 debug: bool=False,
                 log: bool=False,
                 simlogname: str=None,
                 backend: RDDLSimulator=RDDLSimulator,
                 backend_kwargs: Dict={},
                 seeds: list=None):
        '''Creates a new gym environment from the given RDDL domain + instance.
        
        :param domain: the RDDL domain
        :param instance: the RDDL instance
        :param enforce_action_constraints: whether to raise an exception if the
        action constraints are violated
        :param enforce_action_count_non_bool: whether to include non-bool actions
        in check that number of nondef actions don't exceed max-nondef-actions
        :param debug: whether to log compilation information to a log file
        :param log: whether to log simulation data to file
        :param backend: the subclass of RDDLSimulator to use as backend for
        simulation (currently supports numpy and Jax)
        :param backend_kwargs: dictionary of additional named arguments to
        pass to backend (must not include logger)
        :param seeds: list of seeds for the cyclic iterator. Will be seeded in the reset function.
        '''
        super(RDDLEnv, self).__init__()
        self.domain_text = domain
        self.instance_text = instance
        self.enforce_action_constraints = enforce_action_constraints
        self.enforce_action_count_non_bool = enforce_action_count_non_bool

        # time budget for applications limiting time on episodes.
        # hardcoded so cannot be changed externally for the purpose of the competition.
        # TODO: add it to the API after the competition
        self.budget = 240
        self.seeds = RDDLSeeder(a1=1)

        # read and parse domain and instance
        reader = RDDLReader(domain, instance)
        domain = reader.rddltxt

        # parse RDDL file
        parser = RDDLParser(lexer=None, verbose=False)
        parser.build()
        rddl = parser.parse(domain)
        self.model = RDDLLiftedModel(rddl)
        
        # for logging
        ast = self.model._AST
        self.trial = 0
        log_fname = f'{ast.domain.name}_{ast.instance.name}'
        logger = Logger(f'{log_fname}_debug.log') if debug else None
        self.simlogger = None
        if log:
            curpath = os.path.abspath(__file__)
            for _ in range(3):
                curpath = os.path.split(curpath)[0]
            dir = os.path.join(curpath, 'Logs', simlogname, ast.domain.name)
            if not os.path.exists(dir):
                try:
                    os.makedirs(dir)
                except Exception as e:
                    if not isinstance(e, FileExistsError):
                        raise RDDLLogFolderError('Could not create log folder for domain ' + ast.domain.name + ' of method ' + simlogname + ' at path: ' + dir)

            simlog_fname = os.path.join(dir, ast.instance.name)
            self.simlogger = SimLogger(f'{simlog_fname}_log.csv')
        # self.simlogger = SimLogger(f'{log_fname}_log.csv') if log else None
        if self.simlogger:
            self.simlogger.clear(overwrite=False)
        
        # define the model sampler and bounds    
        self.sampler = backend(self.model, logger=logger, **backend_kwargs)
        bounds = RDDLConstraints(self.sampler).bounds

        # set roll-out parameters
        self.horizon = self.model.horizon
        self.discount = self.model.discount
        self.max_allowed_actions = self.model.max_allowed_actions
            
        self.currentH = 0
        self.done = False

        # set default actions
        self.defaultAction = self.model.groundactions()

        # define the actions bounds
        self.actionsranges = self.model.groundactionsranges()
        action_space = Dict()
        for act in self.defaultAction:
            act_range = self.actionsranges[act]
            if act_range in self.model.enums:
                action_space[act] = Discrete(len(self.model.objects[act_range]))            
            elif act_range == 'real':
                action_space[act] = Box(low=bounds[act][0],
                                        high=bounds[act][1],
                                        dtype=np.float32)
            elif act_range == 'bool':
                action_space[act] = Discrete(2)
            elif act_range == 'int':
                high = bounds[act][1]
                if high == np.inf:
                    high = np.iinfo(np.int32).max
                low = bounds[act][0]
                if low == -np.inf:
                    low = np.iinfo(np.int32).min
                action_space[act] = Discrete(int(high - low + 1), start=int(low))
            else:
                raise RDDLTypeError(
                    f'Unknown action value type <{act_range}> in environment.')
        self.action_space = action_space

        # define the states bounds
        if self.sampler.isPOMDP:
            search_dict = self.model.groundobserv()
            ranges = self.model.groundobservranges()
        else:
            search_dict = self.model.groundstates()
            ranges = self.model.groundstatesranges()
            
        state_space = Dict()
        for state in search_dict:
            state_range = ranges[state]
            if state_range in self.model.enums:
                state_space[state] = Discrete(len(self.model.objects[state_range]))          
            elif state_range == 'real':
                state_space[state] = Box(low=bounds[state][0],
                                         high=bounds[state][1],
                                         dtype=np.float32)
            elif state_range == 'bool':
                state_space[state] = Discrete(2)
            elif state_range == 'int':
                high = bounds[state][1]
                if high == np.inf:
                    high = np.iinfo(np.int32).max
                low = bounds[state][0]
                if low == -np.inf:
                    low = np.iinfo(np.int32).min
                state_space[state] = Discrete(int(high - low + 1), start=int(low))
            else:
                raise RDDLTypeError(
                    f'Unknown state value type <{state_range}> in environment.')
        self.observation_space = state_space

        # set the visualizer
        # the next line should be changed for the default behaviour - TextVix
        self._visualizer = TextVisualizer(self.model)
        self._movie_generator = None
        self.state = None
        self.image = None
        self.window = None
        self.to_render = False
        self.image_size = None
    
    def seed(self, seed=None):
        # super(RDDLEnv, self).seed(seed)
        self.sampler.seed(seed)
        return [seed]
    
    def set_visualizer(self, viz, movie_gen=None, movie_per_episode=False, **viz_kwargs):
        if viz is not None:
            self._visualizer = viz(self.model, **viz_kwargs)
        self._movie_generator = movie_gen
        self._movie_per_episode = movie_per_episode
        self._movies = 0
        self.to_render = False

    def step(self, actions):
        if self.done:
            return self.state, 0.0, self.done, {}

        # make sure the action length is of currect size
        if self.enforce_action_count_non_bool:
            action_length = len(actions)
        else:
            action_length = len([
                action 
                for action in actions 
                if self.model.groundactionsranges()[action] == 'bool'
            ])
        if (action_length > self.max_allowed_actions):
            raise RDDLInvalidNumberOfArgumentsError(
                f'Invalid action, expected at most '
                f'{self.max_allowed_actions} entries, '
                f'but got {action_length}.')
        
        # set full action vector
        # values are clipped to be inside the feasible action space
        clipped_actions = copy.deepcopy(self.defaultAction)
        for act in actions:
            if str(self.action_space[act]) == 'Discrete(2)':
                if self.actionsranges[act] == 'bool':
                    clipped_actions[act] = bool(actions[act])
            else:
                clipped_actions[act] = actions[act]
                
        # check action constraints
        if self.enforce_action_constraints:
            self.sampler.check_action_preconditions(clipped_actions)
        
        # sample next state and reward
        obs, reward, self.done = self.sampler.step(clipped_actions)
        state = self.sampler.states
            
        # check if the state invariants are satisfied
        if not self.done:
            self.sampler.check_state_invariants()               

        # log to file
        if self.simlogger is not None:
            self.simlogger.log(
                obs, clipped_actions, reward, self.done, self.currentH)
        
        # update step horizon
        self.currentH += 1
        if self.currentH == self.horizon:
            self.done = True

        # for visualization purposes
        self.state = state

        return obs, reward, self.done, {}

    def reset(self, seed=None):
        self.total_reward = 0
        self.currentH = 0
        obs, self.done = self.sampler.reset()
        self.state = self.sampler.states

        image = self._visualizer.render(self.state)
        if self._movie_generator is not None:
            if self._movie_per_episode:
                self._movie_generator.save_animation(
                    self._movie_generator.env_name + '_' + str(self._movies))
                self._movies += 1
            self._movie_generator.save_frame(image)            
        self.image_size = image.size

        if seed is not None:
            self.seed(seed)
        else:
            seed = self.seeds.Next()
            if seed is not None:
                self.seed(seed)

        # Logging
        if self.simlogger:
            self.trial += 1
            text = '######################################################\n'
            if seed is not None:
                text += f'New Trial, seed={seed}\n'
            else:
                text += f'New Trial\n'
            text += '######################################################'
            self.simlogger.log_free(text)

        return obs

    def pilImageToSurface(self, pilImage):
        return pygame.image.fromstring(
            pilImage.tobytes(), pilImage.size, pilImage.mode).convert()

    def render(self, to_display=True):
        if self._visualizer is not None:
            image = self._visualizer.render(self.state)
            if to_display:
                if not self.to_render:
                    self.to_render = True
                    pygame.init()
                    self.window = pygame.display.set_mode(
                        (self.image_size[0], self.image_size[1]))
                self.window.fill(0)
                pygameSurface = self.pilImageToSurface(image)
                self.window.blit(pygameSurface, (0, 0))
                pygame.display.flip()
    
            if self._movie_generator is not None:
                self._movie_generator.save_frame(image)
    
        return image
    
    def close(self):
        if self.simlogger:
            self.simlogger.close()
                        
        if self.to_render:
            pygame.display.quit()
            pygame.quit()
    
            if self._movie_generator is not None:
                self._movie_generator.save_animation(
                    self._movie_generator.env_name + '_' + str(self._movies))
                self._movies += 1

    @property
    def numConcurrentActions(self):
        return self.max_allowed_actions
    
    @property
    def non_fluents(self):
        return self.model.groundnonfluents()

    @property
    def Budget(self):
        return self.budget



from abc import ABCMeta, abstractmethod
import random

import gym


class BaseAgent(metaclass=ABCMeta):

    @abstractmethod
    def sample_action(self, state):
        pass


class RandomAgent(BaseAgent):
    def __init__(self, action_space, num_actions=1, seed=None):
        self.action_space = action_space
        self.num_actions = num_actions
        self.rng = random.Random(seed)
        if seed is not None:
            self.action_space.seed(seed)

    def sample_action(self, state=None):
        s = self.action_space.sample()
        action = {}
        selected_actions = self.rng.sample(list(s), self.num_actions)
        for sample in selected_actions:
            if isinstance(self.action_space[sample], gym.spaces.Box):
                action[sample] = s[sample][0].item()
            elif isinstance(self.action_space[sample], gym.spaces.Discrete):
                action[sample] = s[sample]
                # if str(self.action_space[sample]) == 'Discrete(2)':
                #     action[sample] = bool(s[sample])
                # else:
                #     action[sample] = s[sample]
        return action


class NoOpAgent(BaseAgent):
    def __init__(self, action_space, num_actions=0):
        self.action_space = action_space
        self.num_actions = num_actions

    def sample_action(self, state=None):
        action = {}
        return action

